from sdm_modbus import meter


class ESPP1(meter.Meter):

    def __init__(self, *args, **kwargs):
        self.model = "ESP-P1-MODBUS"

        super().__init__(*args, **kwargs)

        self.registers = {
            "import_energy_active_low": (0x00, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Imported Energy (Active), Low Tariff", "Wh", 1, 1),
            "import_energy_active_high": (0x02, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Imported Energy (Active), High Tariff", "Wh", 1, 1),
            "export_energy_active_low": (0x04, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Exported Energy (Active), Low Tariff", "Wh", 1, 1),
            "export_energy_active_high": (0x06, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Exported Energy (Active), High Tariff", "Wh", 1, 1),
            "tariff": (0x08, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Current Tariff", [
                "Undefined", "Low", "High"], 1, 1),
            "import_power_active": (0x0A, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Import Power (Active)", "W", 1, 1),
            "export_power_active": (0x0C, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Export Power (Active)", "W", 1, 1),
            "power_failures": (0x0E, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Number of power failures", "", 1, 1),
            "power_failures_long": (0x10, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Number of long power failures", "", 1, 1),
            "l1_voltage_sags": (0x12, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L1 Voltage Sags", "", 1, 1),
            "l2_voltage_sags": (0x14, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L2 Voltage Sags", "", 1, 1),
            "l3_voltage_sags": (0x16, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L3 Voltage Sags", "", 1, 1),
            "l1_voltage_swells": (0x18, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L1 Voltage Swells", "", 1, 1),
            "l2_voltage_swells": (0x1A, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L2 Voltage Swells", "", 1, 1),
            "l3_voltage_swells": (0x1C, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L3 Voltage Swells", "", 1, 1),
            "l1_voltage": (0x1E, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L1 Voltage", "V", 1, .001),
            "l2_voltage": (0x20, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L2 Voltage", "V", 1, .001),
            "l3_voltage": (0x22, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L3 Voltage", "V", 1, .001),
            "l1_current": (0x24, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L1 Current", "A", 1, .001),
            "l2_current": (0x26, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L2 Current", "A", 1, .001),
            "l3_current": (0x28, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L3 Current", "A", 1, .001),
            "l1_import_power_active": (0x2A, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L1 Import Power (Active)", "W", 1, 1),
            "l2_import_power_active": (0x2C, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L2 Import Power (Active)", "W", 1, 1),
            "l3_import_power_active": (0x2E, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L3 Import Power (Active)", "W", 1, 1),
            "l1_export_power_active": (0x30, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L1 Export Power (Active)", "W", 1, 1),
            "l2_export_power_active": (0x32, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L2 Export Power (Active)", "W", 1, 1),
            "l3_export_power_active": (0x34, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "L3 Export Power (Active)", "W", 1, 1),
            "import_gas_volume": (0x36, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Imported Gas Volume", "m3", 1, .001),
            "import_thermal_measurement": (0x38, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Imported Thermal Measurement", "GJ", 1, .001),
            "import_water_volume": (0x3A, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Imported Water Volume", "m3", 1, .001),
            "import_slave_measurement": (0x3C, 2, meter.registerType.HOLDING, meter.registerDataType.UINT32, int, "Imported Slave Measurement", "", 1, 1)
        }



import enum
import time

from pymodbus.constants import Endian
from pymodbus.client.sync import ModbusTcpClient
from pymodbus.client.sync import ModbusSerialClient
from pymodbus.payload import BinaryPayloadBuilder
from pymodbus.payload import BinaryPayloadDecoder
from pymodbus.register_read_message import ReadInputRegistersResponse
from pymodbus.register_read_message import ReadHoldingRegistersResponse


class connectionType(enum.Enum):
    RTU = 1
    TCP = 2


class registerType(enum.Enum):
    INPUT = 1
    HOLDING = 2


class registerDataType(enum.Enum):
    BITS = 1
    UINT8 = 2
    UINT16 = 3
    UINT32 = 4
    UINT64 = 5
    INT8 = 6
    INT16 = 7
    INT32 = 8
    INT64 = 9
    FLOAT16 = 10
    FLOAT32 = 11
    STRING = 12


RETRIES = 3
TIMEOUT = 1
UNIT = 1


class Meter:
    model = "Generic"
    registers = {}

    stopbits = 1
    parity = "N"
    baud = 38400

    wordorder = Endian.Big
    byteorder = Endian.Big

    def __init__(self, **kwargs):
        parent = kwargs.get("parent")

        if parent:
            self.client = parent.client
            self.mode = parent.mode
            self.timeout = parent.timeout
            self.retries = parent.retries

            unit = kwargs.get("unit")

            if unit:
                self.unit = unit
            else:
                self.unit = parent.unit

            if self.mode is connectionType.RTU:
                self.device = parent.device
                self.stopbits = parent.stopbits
                self.parity = parent.parity
                self.baud = parent.baud
            elif self.mode is connectionType.TCP:
                self.host = parent.host
                self.port = parent.port
            else:
                raise NotImplementedError(self.mode)
        else:
            self.timeout = kwargs.get("timeout", TIMEOUT)
            self.retries = kwargs.get("retries", RETRIES)
            self.unit = kwargs.get("unit", UNIT)

            device = kwargs.get("device")

            if device:
                self.device = device

                stopbits = kwargs.get("stopbits")

                if stopbits:
                    self.stopbits = stopbits

                parity = kwargs.get("parity")

                if (parity
                        and parity.upper() in ["N", "E", "O"]):
                    self.parity = parity.upper()
                else:
                    self.parity = False

                baud = kwargs.get("baud")

                if baud:
                    self.baud = baud

                self.mode = connectionType.RTU
                self.client = ModbusSerialClient(
                    method="rtu",
                    port=self.device,
                    stopbits=self.stopbits,
                    parity=self.parity,
                    baudrate=self.baud,
                    timeout=self.timeout
                )
            else:
                self.host = kwargs.get("host")
                self.port = kwargs.get("port", 502)
                self.mode = connectionType.TCP

                self.client = ModbusTcpClient(
                    host=self.host,
                    port=self.port,
                    timeout=self.timeout
                )

    def __repr__(self):
        if self.mode == connectionType.RTU:
            return f"{self.model}({self.device}, {self.mode}: stopbits={self.stopbits}, parity={self.parity}, baud={self.baud}, timeout={self.timeout}, retries={self.retries}, unit={hex(self.unit)})"
        elif self.mode == connectionType.TCP:
            return f"{self.model}({self.host}:{self.port}, {self.mode}: timeout={self.timeout}, retries={self.retries}, unit={hex(self.unit)})"
        else:
            return f"<{self.__class__.__module__}.{self.__class__.__name__} object at {hex(id(self))}>"

    def _read_input_registers(self, address, length):
        for i in range(self.retries):
            if not self.connected():
                self.connect()
                time.sleep(0.1)
                continue

            result = self.client.read_input_registers(address=address, count=length, unit=self.unit)

            if not isinstance(result, ReadInputRegistersResponse):
                continue
            if len(result.registers) != length:
                continue

            return BinaryPayloadDecoder.fromRegisters(result.registers, byteorder=self.byteorder, wordorder=self.wordorder)

        return None

    def _read_holding_registers(self, address, length):
        for i in range(self.retries):
            if not self.connected():
                self.connect()
                time.sleep(0.1)
                continue

            result = self.client.read_holding_registers(address=address, count=length, unit=self.unit)

            if not isinstance(result, ReadHoldingRegistersResponse):
                continue
            if len(result.registers) != length:
                continue

            return BinaryPayloadDecoder.fromRegisters(result.registers, byteorder=self.byteorder, wordorder=self.wordorder)

        return None

    def _write_holding_register(self, address, value):
        return self.client.write_registers(address=address, values=value, unit=self.unit)

    def _encode_value(self, data, dtype):
        builder = BinaryPayloadBuilder(byteorder=self.byteorder, wordorder=self.wordorder)

        try:
            if dtype == registerDataType.FLOAT32:
                builder.add_32bit_float(data)
            elif dtype == registerDataType.INT32:
                builder.add_32bit_int(data)
            elif dtype == registerDataType.UINT32:
                builder.add_32bit_uint(data)
            elif dtype == registerDataType.INT16:
                builder.add_16bit_int(data)
            else:
                raise NotImplementedError(dtype)
        except NotImplementedError:
            raise

        return builder.to_registers()

    def _decode_value(self, data, length, dtype, vtype):
        try:
            if dtype == registerDataType.FLOAT32:
                return vtype(data.decode_32bit_float())
            elif dtype == registerDataType.INT32:
                return vtype(data.decode_32bit_int())
            elif dtype == registerDataType.UINT32:
                return vtype(data.decode_32bit_uint())
            elif dtype == registerDataType.INT16:
                return vtype(data.decode_16bit_int())
            else:
                raise NotImplementedError(dtype)
        except NotImplementedError:
            raise

    def _read(self, value):
        address, length, rtype, dtype, vtype, label, fmt, batch, sf = value

        try:
            if rtype == registerType.INPUT:
                return self._decode_value(self._read_input_registers(address, length), length, dtype, vtype)
            elif rtype == registerType.HOLDING:
                return self._decode_value(self._read_holding_registers(address, length), length, dtype, vtype)
            else:
                raise NotImplementedError(rtype)
        except NotImplementedError:
            raise

    def _read_all(self, values, rtype):
        addr_min = False
        addr_max = False

        for k, v in values.items():
            v_addr = v[0]
            v_length = v[1]

            if addr_min is False:
                addr_min = v_addr
            if addr_max is False:
                addr_max = v_addr + v_length

            if v_addr < addr_min:
                addr_min = v_addr
            if (v_addr + v_length) > addr_max:
                addr_max = v_addr + v_length

        results = {}
        offset = addr_min
        length = addr_max - addr_min

        try:
            if rtype == registerType.INPUT:
                data = self._read_input_registers(offset, length)
            elif rtype == registerType.HOLDING:
                data = self._read_holding_registers(offset, length)
            else:
                raise NotImplementedError(rtype)

            if not data:
                return results

            for k, v in values.items():
                address, length, rtype, dtype, vtype, label, fmt, batch, sf = v

                if address > offset:
                    skip_bytes = address - offset
                    offset += skip_bytes
                    data.skip_bytes(skip_bytes * 2)

                results[k] = self._decode_value(data, length, dtype, vtype)
                offset += length
        except NotImplementedError:
            raise

        return results

    def _write(self, value, data):
        address, length, rtype, dtype, vtype, label, fmt, batch, sf = value

        try:
            if rtype == registerType.HOLDING:
                return self._write_holding_register(address, self._encode_value(data, dtype))
            else:
                raise NotImplementedError(rtype)
        except NotImplementedError:
            raise

    def connect(self):
        return self.client.connect()

    def disconnect(self):
        self.client.close()

    def connected(self):
        return self.client.is_socket_open()

    def get_scaling(self, key):
        address, length, rtype, dtype, vtype, label, fmt, batch, sf = self.registers[key]
        return sf

    def read(self, key, scaling=False):
        if key not in self.registers:
            raise KeyError(key)

        if scaling:
            return self._read(self.registers[key]) * self.get_scaling(key)
        else:
            return self._read(self.registers[key])

    def write(self, key, data):
        if key not in self.registers:
            raise KeyError(key)

        return self._write(self.registers[key], data / self.get_scaling(key))

    def read_all(self, rtype=registerType.INPUT, scaling=False):
        registers = {k: v for k, v in self.registers.items() if (v[2] == rtype)}
        results = {}

        for batch in range(1, max(len(registers), 2)):
            register_batch = {k: v for k, v in registers.items() if (v[7] == batch)}

            if not register_batch:
                break

            results.update(self._read_all(register_batch, rtype))

        if scaling:
            return {k: v * self.get_scaling(k) for k, v in results.items()}
        else:
            return {k: v for k, v in results.items()}



from sdm_modbus import meter


class SDM(meter.Meter):
    pass


class SDM72(SDM):

    def __init__(self, *args, **kwargs):
        self.model = "SDM72"
        self.baud = 9600

        super().__init__(*args, **kwargs)

        self.registers = {
            "l1_voltage": (0x0000, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Voltage", "V", 1, 1),
            "l2_voltage": (0x0002, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Voltage", "V", 1, 1),
            "l3_voltage": (0x0004, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Voltage", "V", 1, 1),
            "l1_current": (0x0006, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Current", "A", 1, 1),
            "l2_current": (0x0008, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Current", "A", 1, 1),
            "l3_current": (0x000a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Current", "A", 1, 1),
            "l1_power_active": (0x000c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Power (Active)", "W", 1, 1),
            "l2_power_active": (0x000e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Power (Active)", "W", 1, 1),
            "l3_power_active": (0x0010, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Power (Active)", "W", 1, 1),
            "l1_power_apparent": (0x0012, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Power (Apparent)", "VA", 1, 1),
            "l2_power_apparent": (0x0014, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Power (Apparent)", "VA", 1, 1),
            "l3_power_apparent": (0x0016, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Power (Apparent)", "VA", 1, 1),
            "l1_power_reactive": (0x0018, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Power (Reactive)", "VAr", 1, 1),
            "l2_power_reactive": (0x001A, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Power (Reactive)", "VAr", 1, 1),
            "l3_power_reactive": (0x001C, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Power (Reactive)", "VAr", 1, 1),
            "l1_power_factor": (0x001e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Power Factor", "", 1, 1),
            "l2_power_factor": (0x0020, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Power Factor", "", 1, 1),
            "l3_power_factor": (0x0022, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Power Factor", "", 1, 1),
            "voltage_ln": (0x002a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L-N Voltage", "V", 1, 1),
            "current_ln": (0x002e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L-N Current", "A", 1, 1),
            "total_line_current": (0x0030, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Line Current", "A", 1, 1),
            "total_power": (0x0034, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Power", "W", 1, 1),
            "total_power_apparent": (0x0038, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Power (Apparent)", "VA", 1, 1),
            "total_power_reactive": (0x003C, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Power (Reactive)", "VAr", 1, 1),
            "total_pf": (0x003E, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Power Factor", "", 1, 1),
            "frequency": (0x0046, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Frequency", "Hz", 1, 1),
            "import_energy_active": (0x0048, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Imported Energy (Active)", "kWh", 1, 1),
            "export_energy_active": (0x004a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Exported Energy (Active)", "kWh", 1, 1),
            "l12_voltage": (0x00c8, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1-L2 Voltage", "V", 2, 1),
            "l23_voltage": (0x00ca, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2-L3 Voltage", "V", 2, 1),
            "l31_voltage": (0x00cc, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3-L1 Voltage", "V", 2, 1),
            "voltage_ll": (0x00ce, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L-L Voltage", "V", 2, 1),
            "neutral_current": (0x00e0, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Neutral Current", "A", 2, 1),
            "total_energy_active": (0x0156, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Energy (Active)", "kWh", 3, 1),
            "total_energy_reactive": (0x0158, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Energy (Reactive)", "kVArh", 3, 1),
            "resettable_total_energy_active": (0x0180, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Resettable Total Energy (Active)", "kWh", 3, 1),
            "resettable_import_enerty_active": (0x0184, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Resettable Import Energy (Active)", "kWh", 3, 1),
            "resettable_export_energy_active": (0x0186, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Resettable Export Energy (Active)", "kWh", 3, 1),
            "net_kwh": (0x018c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Net kWh (Import - Export)", "kWh", 3, 1),
            "import_total_power_active": (0x0500, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Import Power (Active)", "W", 4, 1),
            "export_total_power_active": (0x0502, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Export Power (Active)", "W", 4, 1),

            "demand_time": (0x0000, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Demand Time", "s", 1, 1),
            "demand_period": (0x0002, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Demand Period", "s", 1, 1),
            "system_voltage": (0x0006, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, float, "System Voltage", "V", 1, 1),
            "system_current": (0x0008, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, float, "System Current", "A", 1, 1),
            "system_type": (0x000a, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "System Type", [
                -1, "1P2W", "3P3W", "3P4W"], 1, 1),
            "relay_pulse_width": (0x000c, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Relay Pulse Width", "ms", 1, 1),
            "network_parity_stop": (0x0012, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Network Parity Stop", [
                "N-1", "E-1", "O-1", "N-2"], 1, 1),
            "meter_id": (0x0014, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Meter ID", "", 1, 1),
            "password": (0x0018, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Password", "", 1, 1),
            "baud": (0x001c, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Baud Rate", [
                2400, 4800, 9600, 19200, 38400], 1, 1),
            "system_power": (0x0024, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, float, "System Power", "W", 1, 1),
            "p1_divisor": (0xf910, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "P1 Divisor", [
                "0.001kWh/imp", "0.01kWh/imp", "0.1kWh/imp", "1kWh/imp", "10kWh/imp", "100kWh/imp"], 2, 1)
        }


class SDM120(SDM):

    def __init__(self, *args, **kwargs):
        self.model = "SDM120"
        self.baud = 2400

        super().__init__(*args, **kwargs)

        self.registers = {
            "voltage": (0x0000, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Voltage", "V", 1, 1),
            "current": (0x0006, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Current", "A", 1, 1),
            "power_active": (0x000c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Power (Active)", "W", 1, 1),
            "power_apparent": (0x0012, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Power (Apparent)", "VA", 1, 1),
            "power_reactive": (0x0018, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Power (Reactive)", "VAr", 1, 1),
            "power_factor": (0x001e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Power Factor", "", 1, 1),
            "phase_angle": (0x0024, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Phase Angle", "°", 1, 1),
            "frequency": (0x0046, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Frequency", "Hz", 1, 1),
            "import_energy_active": (0x0048, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Imported Energy (Active)", "kWh", 1, 1),
            "export_energy_active": (0x004a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Exported Energy (Active)", "kWh", 1, 1),
            "import_energy_reactive": (0x004c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Imported Energy (Reactive)", "kVArh", 1, 1),
            "export_energy_reactive": (0x004e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Exported Energy (Reactive)", "kVArh", 1, 1),
            "total_demand_power_active": (0x0054, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Demand Power (Active)", "W", 2, 1),
            "maximum_total_demand_power_active": (0x0056, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Total Demand Power (Active)", "W", 2, 1),
            "import_demand_power_active": (0x0058, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Import Demand Power (Active)", "W", 2, 1),
            "maximum_import_demand_power_active": (0x005a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Import Demand Power (Active)", "W", 2, 1),
            "export_demand_power_active": (0x005c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Export Demand Power (Active)", "W", 2, 1),
            "maximum_export_demand_power_active": (0x005e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Export Demand Power (Active)", "W", 2, 1),
            "total_demand_current": (0x0102, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Demand Current", "A", 3, 1),
            "maximum_total_demand_current": (0x0108, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Total Demand Current", "A", 3, 1),
            "total_energy_active": (0x0156, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Energy (Active)", "kWh", 4, 1),
            "total_energy_reactive": (0x0158, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Energy (Reactive)", "kVArh", 4, 1),

            "demand_time": (0x0000, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Demand Time", "s", 1, 1),
            "demand_period": (0x0002, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Demand Period", "s", 1, 1),
            "relay_pulse_width": (0x000c, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Relay Pulse Width", "ms", 1, 1),
            "network_parity_stop": (0x0012, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Network Parity Stop", [
                "N-1", "E-1", "O-1", "N-2"], 1, 1),
            "meter_id": (0x0014, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Meter ID", "", 1, 1),
            "baud": (0x001c, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Baud Rate", [
                2400, 4800, 9600, -1, -1, 1200], 1, 1),
            "p1_output_mode": (0x0056, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "P1 Output Mode", [
                0x0, "Import Energy (Active)", "Import + Export Energy (Active)", 0x3, "Export Energy (Active)",
                "Import Energy (Reactive)", "Import + Export Energy (Reactive)", 0x7, "Export Energy (Reactive)"], 2, 1),
            "display_scroll_timing": (0xf900, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Display Scroll Timing", "s", 3, 1),
            "p1_divisor": (0xf910, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "P1 Divisor", [
                "0.001kWh/imp", "0.01kWh/imp", "0.1kWh/imp", "1kWh/imp"], 3, 1),
            "measurement_mode": (0xf920, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Measurement Mode", [
                0x0, "Total Imported", "Total Imported + Exported", "Total Imported - Exported"], 3, 1),
            "indicator_mode": (0xf930, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Pulse/LED Indicator Mode", [
                "Import + Export Energy (Active)", "Import Energy (Active)", "Export Energy (Active)"], 3, 1)
        }


class SDM230(SDM):

    def __init__(self, *args, **kwargs):
        self.model = "SDM230"
        self.baud = 2400

        super().__init__(*args, **kwargs)

        self.registers = {
            "voltage": (0x0000, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Voltage", "V", 1, 1),
            "current": (0x0006, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Current", "A", 1, 1),
            "power_active": (0x000c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Power (Active)", "W", 1, 1),
            "power_apparent": (0x0012, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Power (Apparent)", "VA", 1, 1),
            "power_reactive": (0x0018, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Power (Reactive)", "VAr", 1, 1),
            "power_factor": (0x001e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Power Factor", "", 1, 1),
            "phase_angle": (0x0024, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Phase Angle", "°", 1, 1),
            "frequency": (0x0046, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Frequency", "Hz", 1, 1),
            "import_energy_active": (0x0048, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Imported Energy (Active)", "kWh", 1, 1),
            "export_energy_active": (0x004a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Exported Energy (Active)", "kWh", 1, 1),
            "import_energy_reactive": (0x004c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Imported Energy (Reactive)", "kVArh", 1, 1),
            "export_energy_reactive": (0x004e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Exported Energy (Reactive)", "kVArh", 1, 1),
            "total_demand_power_active": (0x0054, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Demand Power (Active)", "W", 2, 1),
            "maximum_total_demand_power_active": (0x0056, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Total Demand Power (Active)", "W", 2, 1),
            "import_demand_power_active": (0x0058, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Import Demand Power (Active)", "W", 2, 1),
            "maximum_import_demand_power_active": (0x005a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Import Demand Power (Active)", "W", 2, 1),
            "export_demand_power_active": (0x005c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Export Demand Power (Active)", "W", 2, 1),
            "maximum_export_demand_power_active": (0x005e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Export Demand Power (Active)", "W", 2, 1),
            "total_demand_current": (0x0102, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Demand Current", "A", 3, 1),
            "maximum_total_demand_current": (0x0108, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Total Demand Current", "A", 3, 1),
            "total_energy_active": (0x0156, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Energy (Active)", "kWh", 4, 1),
            "total_energy_reactive": (0x0158, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Energy (Reactive)", "kVArh", 4, 1),

            "relay_pulse_width": (0x000c, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Relay Pulse Width", "ms", 1, 1),
            "network_parity_stop": (0x0012, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Network Parity Stop", [
                "N-1", "E-1", "O-1", "N-2"], 1, 1),
            "meter_id": (0x0014, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Meter ID", "", 1, 1),
            "baud": (0x001c, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Baud Rate", [
                2400, 4800, 9600, -1, -1, 1200], 1, 1),
            "p1_output_mode": (0x0056, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "P1 Output Mode", [
                0x0, "Import Energy (Active)", "Import + Export Energy (Active)", 0x3, "Export Energy (Active)",
                "Import Energy (Reactive)", "Import + Export Energy (Reactive)", 0x7, "Export Energy (Reactive)"], 2, 1),
            "display_scroll_timing": (0xf900, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Display Scroll Timing", "s", 3, 1),
            "p1_divisor": (0xf910, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "P1 Divisor", [
                "0.001kWh/imp", "0.01kWh/imp", "0.1kWh/imp", "1kWh/imp"], 3, 1),
            "measurement_mode": (0xf920, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Measurement Mode", [
                0x0, "Total Imported", "Total Imported + Exported", "Total Imported - Exported"], 3, 1),
            "running_time": (0xf930, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Running Time", "h", 3, 1)
        }


class SDM630(SDM):

    def __init__(self, *args, **kwargs):
        self.model = "SDM630"
        self.baud = 9600

        super().__init__(*args, **kwargs)

        self.registers = {
            "l1_voltage": (0x0000, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Voltage", "V", 1, 1),
            "l2_voltage": (0x0002, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Voltage", "V", 1, 1),
            "l3_voltage": (0x0004, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Voltage", "V", 1, 1),
            "l1_current": (0x0006, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Current", "A", 1, 1),
            "l2_current": (0x0008, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Current", "A", 1, 1),
            "l3_current": (0x000a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Current", "A", 1, 1),
            "l1_power_active": (0x000c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Power (Active)", "W", 1, 1),
            "l2_power_active": (0x000e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Power (Active)", "W", 1, 1),
            "l3_power_active": (0x0010, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Power (Active)", "W", 1, 1),
            "l1_power_apparent": (0x0012, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Power (Apparent)", "VA", 1, 1),
            "l2_power_apparent": (0x0014, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Power (Apparent)", "VA", 1, 1),
            "l3_power_apparent": (0x0016, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Power (Apparent)", "VA", 1, 1),
            "l1_power_reactive": (0x0018, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Power (Reactive)", "VAr", 1, 1),
            "l2_power_reactive": (0x001A, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Power (Reactive)", "VAr", 1, 1),
            "l3_power_reactive": (0x001C, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Power (Reactive)", "VAr", 1, 1),
            "l1_power_factor": (0x001e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Power Factor", "", 1, 1),
            "l2_power_factor": (0x0020, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Power Factor", "", 1, 1),
            "l3_power_factor": (0x0022, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Power Factor", "", 1, 1),
            "l1_phase_angle": (0x0024, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Phase Angle", "°", 1, 1),
            "l2_phase_angle": (0x0026, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Phase Angle", "°", 1, 1),
            "l3_phase_angle": (0x0028, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Phase Angle", "°", 1, 1),
            "voltage_ln": (0x002a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L-N Voltage", "V", 1, 1),
            "current_ln": (0x002e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L-N Current", "A", 1, 1),
            "total_line_current": (0x0030, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Line Current", "A", 1, 1),
            "total_power_active": (0x0034, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Power (Active)", "W", 1, 1),
            "total_power_apparent": (0x0038, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Power (Apparent)", "VA", 1, 1),
            "total_power_reactive": (0x003C, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Power (Reactive)", "VAr", 1, 1),
            "total_power_factor": (0x003E, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Power Factor", "", 1, 1),
            "total_phase_angle": (0x0042, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Phase Angle", "°", 1, 1),
            "frequency": (0x0046, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Frequency", "Hz", 1, 1),
            "import_energy_active": (0x0048, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Imported Energy (Active)", "kWh", 1, 1),
            "export_energy_active": (0x004a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Exported Energy (Active)", "kWh", 1, 1),
            "import_energy_reactive": (0x004c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Imported Energy (Reactive)", "kVArh", 1, 1),
            "export_energy_reactive": (0x004e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Exported Energy (Reactive)", "kVArh", 1, 1),
            "total_energy_apparent": (0x0050, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Energy (Apparent)", "kVAh", 2, 1),
            "total_current": (0x0052, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Current", "A", 2, 1),
            "total_import_demand_power_active": (0x0054, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Import Demand Power (Active)", "W", 2, 1),
            "maximum_import_demand_power_apparent": (0x0056, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Import Demand Power (Apparent)", "VA", 2, 1),
            "import_demand_power_active": (0x0058, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Import Demand Power (Active)", "W", 2, 1),
            "maximum_import_demand_power_active": (0x005a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Import Demand Power (Active)", "W", 2, 1),
            "export_demand_power_active": (0x005c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Export Demand Power (Active)", "W", 2, 1),
            "maximum_export_demand_power_active": (0x005e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Export Demand Power (Active)", "W", 2, 1),
            "total_demand_power_apparent": (0x0064, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Demand Power (Apparent)", "VA", 2, 1),
            "maximum_demand_power_apparent": (0x0066, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum System Power (Apparent)", "VA", 2, 1),
            "neutral_demand_current": (0x0068, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Neutral Demand Current", "A", 2, 1),
            "maximum_neutral_demand_current": (0x006a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum Neutral Demand Current", "A", 2, 1),
            "l12_voltage": (0x00c8, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1-L2 Voltage", "V", 3, 1),
            "l23_voltage": (0x00ca, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2-L3 Voltage", "V", 3, 1),
            "l31_voltage": (0x00cc, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3-L1 Voltage", "V", 3, 1),
            "voltage_ll": (0x00ce, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L-L Voltage", "V", 3, 1),
            "neutral_current": (0x00e0, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Neutral Current", "A", 3, 1),
            "l1n_voltage_thd": (0x00ea, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1-N Voltage THD", "%", 3, 1),
            "l2n_voltage_thd": (0x00ec, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2-N Voltage THD", "%", 3, 1),
            "l3n_voltage_thd": (0x00ee, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3-N Voltage THD", "%", 3, 1),
            "l1_current_thd": (0x00f0, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Current THD", "%", 3, 1),
            "l2_current_thd": (0x00f2, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Current THD", "%", 3, 1),
            "l3_current_thd": (0x00f4, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Current THD", "%", 3, 1),
            "voltage_ln_thd": (0x00f8, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L-N Voltage THD", "%", 3, 1),
            "current_thd": (0x00fa, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Current THD", "%", 3, 1),
            "total_pf": (0x00fe, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Power Factor", "", 3, 1),
            "l1_demand_current": (0x0102, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Demand Current", "A", 3, 1),
            "l2_demand_current": (0x0104, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Demand Current", "A", 3, 1),
            "l3_demand_current": (0x0106, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Demand Current", "A", 3, 1),
            "maximum_l1_demand_current": (0x0108, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum L1 Demand Current", "A", 3, 1),
            "maximum_l2_demand_current": (0x010a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum L2 Demand Current", "A", 3, 1),
            "maximum_l3_demand_current": (0x010c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Maximum L3 Demand Current", "A", 3, 1),
            "l12_voltage_thd": (0x014e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1-L2 Voltage THD", "%", 4, 1),
            "l23_voltage_thd": (0x0150, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2-L3 Voltage THD", "%", 4, 1),
            "l31_voltage_thd": (0x0152, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3-L1 Voltage THD", "%", 4, 1),
            "voltage_ll_thd": (0x0154, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L-L Voltage THD", "%", 4, 1),
            "total_energy_active": (0x0156, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Energy (Active)", "kWh", 4, 1),
            "total_energy_reactive": (0x0158, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "Total Energy (Reactive)", "kVArh", 4, 1),
            "l1_import_energy_active": (0x015a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Import Energy (Active)", "kWh", 4, 1),
            "l2_import_energy_active": (0x015c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Import Energy (Active)", "kWh", 4, 1),
            "l3_import_energy_active": (0x015e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Import Energy (Active)", "kWh", 4, 1),
            "l1_export_energy_active": (0x0160, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Export Energy (Active)", "kWh", 4, 1),
            "l2_export_energy_active": (0x0162, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Export Energy (Active)", "kWh", 4, 1),
            "l3_export_energy_active": (0x0164, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Export Energy (Active)", "kWh", 4, 1),
            "l1_energy_active": (0x0166, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Total Energy (Active)", "kWh", 4, 1),
            "l2_energy_active": (0x0168, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Total Energy (Active)", "kWh", 4, 1),
            "l3_energy_active": (0x016a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Total Energy (Active)", "kWh", 4, 1),
            "l1_import_energy_reactive": (0x016c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Import Energy (Reactive)", "kVArh", 4, 1),
            "l2_import_energy_reactive": (0x016e, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Import Energy (Reactive)", "kVArh", 4, 1),
            "l3_import_energy_reactive": (0x0170, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Import Energy (Reactive)", "kVArh", 4, 1),
            "l1_export_energy_reactive": (0x0172, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Export Energy (Reactive)", "kVArh", 4, 1),
            "l2_export_energy_reactive": (0x0174, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Export Energy (Reactive)", "kVArh", 4, 1),
            "l3_export_energy_reactive": (0x0176, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Export Energy (Reactive)", "kVArh", 4, 1),
            "l1_energy_reactive": (0x0178, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L1 Total Energy (Reactive)", "kVArh", 4, 1),
            "l2_energy_reactive": (0x017a, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L2 Total Energy (Reactive)", "kVArh", 4, 1),
            "l3_energy_reactive": (0x017c, 2, meter.registerType.INPUT, meter.registerDataType.FLOAT32, float, "L3 Total Energy (Reactive)", "kVArh", 4, 1),

            "demand_time": (0x0000, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Demand Time", "s", 1, 1),
            "demand_period": (0x0002, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Demand Period", "s", 1, 1),
            "system_voltage": (0x0006, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, float, "System Voltage", "V", 1, 1),
            "system_current": (0x0008, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, float, "System Current", "A", 1, 1),
            "system_type": (0x000a, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "System Type", [
                -1, "1P2W", "3P3W", "3P4W"], 1, 1),
            "relay_pulse_width": (0x000c, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Relay Pulse Width", "ms", 1, 1),
            "network_parity_stop": (0x0012, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Network Parity Stop", [
                "N-1", "E-1", "O-1", "N-2"], 1, 1),
            "meter_id": (0x0014, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Meter ID", "", 1, 1),
            "baud": (0x001c, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "Baud Rate", [
                2400, 4800, 9600, 19200, 38400], 1, 1),
            "system_power": (0x0024, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, float, "System Power", "W", 1, 1),
            "p1_divisor": (0xf910, 2, meter.registerType.HOLDING, meter.registerDataType.FLOAT32, int, "P1 Divisor", [
                "0.001kWh/imp", "0.01kWh/imp", "0.1kWh/imp", "1kWh/imp", "10kWh/imp", "100kWh/imp"], 2, 1)
        }



from sdm_modbus import meter
from pymodbus.constants import Endian


class GARO(meter.Meter):
    pass


class GNM3D(GARO):

    def __init__(self, *args, **kwargs):
        self.model = "GNM3D"
        self.wordorder = Endian.Little

        super().__init__(*args, **kwargs)

        self.registers = {
            "l1n_voltage": (0x0000, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L1-N Voltage", "V", 1, 0.1),
            "l2n_voltage": (0x0002, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L2-N Voltage", "V", 1, 0.1),
            "l3n_voltage": (0x0004, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L3-N Voltage", "V", 1, 0.1),
            "l12_voltage": (0x0006, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L1-L2 Voltage", "V", 1, 0.1),
            "l23_voltage": (0x0008, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L2-L3 Voltage", "V", 1, 0.1),
            "l31_voltage": (0x000A, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L3-L1 Voltage", "V", 1, 0.1),
            "l1_current": (0x000C, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L1 Current", "A", 1, 0.001),
            "l2_current": (0x000E, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L2 Current", "A", 1, 0.001),
            "l3_current": (0x0010, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L3 Current", "A", 1, 0.001),
            "l1_power_active": (0x0012, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L1 Power (Active)", "W", 1, 0.1),
            "l2_power_active": (0x0014, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L2 Power (Active)", "W", 1, 0.1),
            "l3_power_active": (0x0016, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L3 Power (Active)", "W", 1, 0.1),
            "l1_energy_apparent": (0x0018, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L1 Energy (Apparent)", "VA", 1, 0.1),
            "l2_energy_apparent": (0x001A, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L2 Energy (Apparent)", "VA", 1, 0.1),
            "l3_energy_apparent": (0x001C, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L3 Energy (Apparent)", "VA", 1, 0.1),
            "l1_energy_reactive": (0x001E, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L1 Energy (Reactive)", "VAr", 1, 0.1),
            "l2_energy_reactive": (0x0020, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L2 Energy (Reactive)", "VAr", 1, 0.1),
            "l3_energy_reactive": (0x0022, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L3 Energy (Reactive)", "VAr", 1, 0.1),
            "voltage_ln": (0x0024, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L-N Voltage", "V", 2, 0.1),
            "voltage_ll": (0x0026, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L-L Voltage", "V", 2, 0.1),
            "power_active": (0x0028, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "Power (Active)", "W", 2, 0.1),
            "power_apparent": (0x002A, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "Power (Apparent)", "VA", 2, 0.1),
            "power_reactive": (0x002C, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "Power (Reactive)", "VAr", 2, 0.1),
            "l1_power_factor": (0x002E, 1, meter.registerType.INPUT, meter.registerDataType.INT16, int, "L1 Power Factor", "", 2, 0.001),
            "l2_power_factor": (0x002F, 1, meter.registerType.INPUT, meter.registerDataType.INT16, int, "L2 Power Factor", "", 2, 0.001),
            "l3_power_factor": (0x0030, 1, meter.registerType.INPUT, meter.registerDataType.INT16, int, "L3 Power Factor", "", 2, 0.001),
            "power_factor": (0x0031, 1, meter.registerType.INPUT, meter.registerDataType.INT16, int, "Power Factor", "", 2, 0.001),
            "frequency": (0x0033, 1, meter.registerType.INPUT, meter.registerDataType.INT16, int, "Frequency", "Hz", 2, 0.1),
            "import_energy_active": (0x0034, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "Imported Energy (Active)", "kWh", 2, 0.1),
            "import_energy_reactive": (0x0036, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "Imported Energy (Reactive)", "kVArh", 2, 0.1),
            "demand_power_active": (0x0038, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "Demand Power (Active)", "W", 2, 0.1),
            "maximum_demand_power_active": (0x003A, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "Maximum Demand Power (Active)", "W", 2, 0.1),
            "l1_import_energy_active": (0x0040, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L1 Imported Energy (Active)", "kWh", 2, 0.1),
            "l2_import_energy_active": (0x0042, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L2 Imported Energy (Active)", "kWh", 2, 0.1),
            "l3_import_energy_active": (0x0044, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "L3 Imported Energy (Active)", "kWh", 2, 0.1),
            "export_energy_active": (0x004E, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "Exported Energy (Active)", "kWh", 2, 0.1),
            "export_energy_reactive": (0x0050, 2, meter.registerType.INPUT, meter.registerDataType.INT32, int, "Exported Energy (Reactive)", "kVArh", 2, 0.1)
        }



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sdm_modbus']

package_data = \
{'': ['*']}

install_requires = \
['pymodbus>=2.4.0,<3.0.0']

setup_kwargs = {
    'name': 'sdm-modbus',
    'version': '0.5.2',
    'description': 'Eastron SDM Modbus parser library',
    'long_description': '# sdm_modbus\n\nsdm_modbus is a python library that collects data from Eastron SDM single- and three-phase kWh meters over Modbus RTU or Modbus TCP.\n\nSupported devices:\n* [Eastron SDM72D-M](https://www.eastroneurope.com/products/view/sdm72modbus)\n* [Eastron SDM120-Modbus](https://www.eastroneurope.com/products/view/sdm120modbus)\n* [Eastron SDM230-Modbus](https://www.eastroneurope.com/products/view/sdm230modbus)\n* [Eastron SDM630-Modbus](https://www.eastroneurope.com/products/view/sdm630modbus)\n* [Garo GNM3D-RS485](https://www.garo.se/en/professional/products/installation-products/din-rail-components/energy-meters/energymeter-3p-modbus-rs485)\n* [ESP2866/32 P1 Modbus](https://github.com/nmakel/esp_p1_modbus)\n\n## Installation\n\nTo install, either clone this project and install using `poetry`:\n\n```poetry install```\n\nor install the package from PyPi:\n\n```pip3 install sdm_modbus```\n\n## Usage\n\nThe script `example-tcp.py` provides a minimal example of connecting to and displaying all input and holding registers on a **SDM120** over **Modbus TCP**. To display values as a JSON object, add `--json`.\n\n```\nusage: example-tcp.py [-h] [--unit UNIT] [--json] host port\n\npositional arguments:\n  host         Modbus TCP address\n  port         Modbus TCP port\n\noptional arguments:\n  -h, --help   show this help message and exit\n  --unit UNIT  Modbus device address\n  --json       Output as JSON\n```\n\nOutput:\n\n```\nSDM120(10.0.0.123:502, connectionType.TCP: timeout=1, retries=3, unit=0x1):\n\nInput Registers:\n    Voltage: 237.00V\n    Current: 8.63A\n    Power (Active): -1919.70W\n    Power (Apparent): 2035.57VA\n    Power (Reactive): -676.70VA\n    Power Factor: -0.94\n    Phase Angle: 0.00°\n    Frequency: 50.00Hz\n    Imported Energy (Active): 1551.37kWh\n    Imported Energy (Active): 1335.69kWh\n    Imported Energy (Reactive): 0.01kVAh\n    Exported Energy (Reactive): 4362.14kVAh\n    Total Demand Power (Active): 1668.02W\n    Maximum Total Demand Power (Active): 3347.26W\n    Import Demand Power (Active): 0.00W\n    Maximum Import Demand Power (Active): 3347.26W\n    Export Demand Power (Active): 1668.02W\n    Maximum Export Demand Power (Active): 2109.45W\n    Total Demand Current: 7.57A\n    Maximum Total Demand Current: 14.97A\n    Total Energy (Active): 2887.07kWh\n    Total Energy (Reactive): 0.00kVAh\n\nHolding Registers:\n    Demand Time: 1s\n    Demand Period: 60s\n    Relay Pulse Width: 60ms\n    Network Parity Stop: N-1\n    Meter ID: 1\n    Baud Rate: 9600\n    P1 Output Mode: Export Energy (Active)\n    Display Scroll Timing: 0s\n    P1 Divisor: 0.001kWh/imp\n    Measurement Mode: 0\n    Pulse/LED Indicator Mode: Import + Export Energy (Active)\n```\n\n### Connecting\n\nIf you wish to use Modbus TCP the following parameters are relevant:\n\n`host = IP or DNS name of your Modbus TCP gateway, required`  \n`port = TCP port of the Modbus TCP gateway, required`  \n`unit = Modbus device address, default=1, optional`\n\nWhile if you are using a Modbus RTU connection you can specify:\n\n`device = path to serial device, e.g. /dev/ttyUSB0, required`  \n`baud = baud rate of your SDM unit, defaults to factory default, optional`  \n`unit = Modbus device address, defaults to 1, optional`\n\nConnecting to the meter:\n\n```\n    >>> import sdm_modbus\n\n    # SDM120 over Modbus TCP\n    >>> device = sdm_modbus.SDM120(host="10.0.0.123", port=502)\n\n    # SDM630 over Modbus RTU\n    >>> device = sdm_modbus.SDM630(device="/dev/ttyUSB0", baud=9600)\n```\n\nTest the connection:\n\n```\n    >>> device.connected()\n    True\n```\n\nPrinting the object yields basic device parameters:\n\n```\n    >>> device\n    SDM120(10.0.0.123:502, connectionType.TCP: timeout=1, retries=3, unit=0x1):\n```\n\n### Connecting to Multiple Devices\n\nRe-using an existing RTU or TCP connection is possible by providing an already connected device as `parent` when creating a new instance. This may be necessary if the Modbus TCP gateway only accepts a limited number of connections, or you wish to address multiple RTU devices on the same bus. For example:\n\n```\n    # Connect to a SDM630 over Modbus TCP\n    >>> device_1 = sdm_modbus.SDM630(host="10.0.0.123", port=502, unit=1)\n\n    # Connect to a SDM630 using the existing connection\n    >>> device_2 = sdm_modbus.SDM630(parent=device_1, unit=2)\n```\n\n### Reading Registers\n\nReading a single input register by name:\n\n```\n    >>> device.read("voltage")\n    236.89999389648438\n```\n\nRead all input registers by passing the `sdm_modbus.registerType.INPUT` enum to `read_all()`. Leave this blank to read both `INPUT` and `HOLDING` registers:\n\n```\n    >>> device.read_all(sdm_modbus.registerType.INPUT)\n    {\n        "voltage": 238.60000610351562,\n        "current": 7.59499979019165,\n        "power_active": -1673.800048828125,\n        "power_apparent": 1797.5904541015625,\n        "power_reactive": -655.4000244140625,\n        "power_factor": -0.9311425685882568,\n        "phase_angle": 0.0,\n        "frequency": 50.0,\n        "import_energy_active": 1556.35595703125,\n        "export_energy_active": 1345.9210205078125,\n        "import_energy_reactive": 0.014999999664723873,\n        "export_energy_reactive": 4376.02001953125,\n        "total_demand_power_active": 1659.360107421875,\n        "maximum_total_demand_power_active": 3347.26318359375,\n        "import_demand_power_active": 0.0,\n        "maximum_import_demand_power_active": 3347.26318359375,\n        "export_demand_power_active": 1659.360107421875,\n        "maximum_export_demand_power_active": 2109.4541015625,\n        "total_demand_current": 7.531858921051025,\n        "maximum_total_demand_current": 14.968546867370605,\n        "total_energy_active": 2902.277099609375,\n        "total_energy_reactive": 4376.03515625\n    }\n\n    >>> device.read_all(sdm_modbus.registerType.HOLDING)\n    {\n        "demand_time": 1,\n        "demand_period": 60,\n        "relay_pulse_width": 60,\n        "network_parity_stop": 0,\n        "meter_id": 1,\n        "baud": 2,\n        "p1_output_mode": 4,\n        "display_scroll_timing": 0,\n        "p1_divisor": 0,\n        "measurement_mode": 0,\n        "indicator_mode": 0\n    }\n```\n\n### Writing Registers\n\nWriting to holding registers is also possible. Setting a new baud rate, for example:\n\n```\n    >>> device.write("baud", 2)\n    WriteMultipleRegisterResponse (28,2)\n```\n\nYou will need to **enable setup mode on your device** by pressing the setup button for 5 seconds, otherwise you will receive a `Exception Response(134, 6, GatewayNoResponse)` or similar.\n\n### Register Details\n\nIf you need more information about a particular register, to look up the units or enumerations, for example:\n\n```\n    >>> device.registers["voltage"]\n        # address, length, type, datatype, valuetype, name, unit, batching\n        (\n            0,\n            2,\n            <registerType.INPUT: 1>,\n            <registerDataType.FLOAT32: 11>,\n            <class \'float\'>,\n            \'Voltage\',\n            \'V\',\n            1\n        )\n\n    >>> device.registers["p1_divisor"]\n        # address, length, type, datatype, valuetype, name, unit, batching\n        (\n            63760,\n            2,\n            <registerType.HOLDING: 2>,\n            <registerDataType.FLOAT32: 11>,\n            <class \'int\'>,\n            \'P1 Divisor\',\n            [\'0.001kWh/imp\', \'0.01kWh/imp\', \'0.1kWh/imp\', \'1kWh/imp\'],\n            2\n        )\n```\n\n## Contributing\n\nContributions are more than welcome, especially testing on supported units, and adding other Eastron SDM units.',
    'author': 'nmakel',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nmakel/sdm_modbus',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
